import typing

from fntypes import Nothing, Option, Some

T = typing.TypeVar("T")


def from_optional(value: T | None) -> Option[T]:
    return Some(value) if value is not None else Nothing()


__all__ = ("from_optional",)



import importlib.util as importlib_util
import os
import pathlib
import sys
import typing
from abc import ABC, abstractmethod

from fntypes.option import Option

from telegrinder.api.api import API
from telegrinder.tools.global_context.abc import ABCGlobalContext
from telegrinder.types.objects import Update

T = typing.TypeVar("T")


class PathExistsError(BaseException):
    pass


class ABCDispatch(ABC):
    @property
    @abstractmethod
    def global_context(self) -> ABCGlobalContext:
        pass

    @abstractmethod
    async def feed(self, event: Update, api: API) -> bool:
        pass

    @abstractmethod
    def load(self, external: typing.Self) -> None:
        pass

    @abstractmethod
    def get_view(self, of_type: type[T]) -> Option[T]:
        pass

    def load_many(self, *externals: typing.Self) -> None:
        for external in externals:
            self.load(external)

    def load_from_dir(self, directory: str | pathlib.Path) -> bool:
        """Loads dispatchers from a directory containing Python modules where global variables
        are declared with instances of dispatch.
        Returns True if dispatchers were found, otherwise False."""

        directory = pathlib.Path(directory)

        if not directory.exists():
            raise PathExistsError(f"Path {str(directory)!r} does not exists.")

        dps: list[typing.Self] = []
        for root, _, files in os.walk(directory):
            for f in files:
                if f.endswith(".py") and f != "__init__.py":
                    module_path = os.path.join(root, f)
                    module_name = os.path.splitext(os.path.relpath(module_path, directory))[0]
                    module_name = module_name.replace(os.sep, ".")

                    spec = importlib_util.spec_from_file_location(module_name, module_path)
                    if spec is None or spec.loader is None:
                        continue

                    module = importlib_util.module_from_spec(spec)
                    sys.modules[module_name] = module
                    spec.loader.exec_module(module)

                    for obj in module.__dict__.values():
                        if isinstance(obj, self.__class__):
                            dps.append(obj)

        self.load_many(*dps)
        return bool(dps)


__all__ = ("ABCDispatch",)



import typing
from abc import ABC, abstractmethod

from fntypes.result import Result

from telegrinder.api import API
from telegrinder.bot.dispatch.context import Context

Event = typing.TypeVar("Event")
Handler = typing.Callable[..., typing.Awaitable[typing.Any]]


class ABCErrorHandler(ABC, typing.Generic[Event]):
    @abstractmethod
    def __call__(
        self,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> typing.Callable[[typing.Callable[..., typing.Any]], typing.Callable[..., typing.Any]]:
        """Decorator for registering callback as a catcher for the error handler."""

    @abstractmethod
    async def run(
        self,
        handler: Handler,
        event: Event,
        api: API,
        ctx: Context,
    ) -> Result[typing.Any, typing.Any]:
        """Run the error handler."""


__all__ = ("ABCErrorHandler",)



from fntypes.option import Option

from telegrinder.bot.dispatch.view import BaseStateView
from telegrinder.bot.dispatch.waiter_machine.hasher.hasher import ECHO, Event, Hasher
from telegrinder.tools.functional import from_optional


class StateViewHasher(Hasher[Event, int]):
    view: BaseStateView[Event]

    def __init__(self, view: BaseStateView[Event]) -> None:
        self.view = view
        super().__init__(view.__class__, get_hash_from_data=ECHO)

    def get_data_from_event(self, event: Event) -> Option[int]:
        return from_optional(self.view.get_state_key(event))


__all__ = ("StateViewHasher",)



import dataclasses
import typing

from fntypes.result import Error, Ok, Result

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.modules import logger
from telegrinder.node.base import is_node
from telegrinder.tools.error_handler.abc import ABCErrorHandler, Event, Handler
from telegrinder.tools.error_handler.error import CatcherError
from telegrinder.tools.magic import get_annotations, magic_bundle

F = typing.TypeVar("F", bound="FuncCatcher[typing.Any]")
ExceptionT = typing.TypeVar("ExceptionT", bound=BaseException, contravariant=True)
FuncCatcher = typing.Callable[typing.Concatenate[ExceptionT, ...], typing.Awaitable[typing.Any]]


async def run_handler(
    handler: Handler,
    event: typing.Any,
    ctx: dict[str, typing.Any],
) -> typing.Any:
    annotations = tuple(get_annotations(handler).values())
    start_idx = 0 if is_node(None if not annotations else annotations[0]) else 1
    context = magic_bundle(handler, ctx, start_idx=start_idx)
    return await (handler(event, **context) if start_idx == 1 else handler(**context))


@dataclasses.dataclass(frozen=True, repr=False, slots=True)
class Catcher(typing.Generic[Event]):
    func: FuncCatcher[BaseException]
    exceptions: list[type[BaseException] | BaseException] = dataclasses.field(
        default_factory=lambda: [],
        kw_only=True,
    )
    logging: bool = dataclasses.field(default=False, kw_only=True)
    raise_exception: bool = dataclasses.field(default=False, kw_only=True)
    ignore_errors: bool = dataclasses.field(default=False, kw_only=True)

    def __repr__(self) -> str:
        return "<Catcher: function={!r}, logging={}, raise_exception={}, ignore_errors={}>".format(
            self.func.__name__,
            self.logging,
            self.raise_exception,
            self.ignore_errors,
        )

    async def __call__(
        self,
        handler: Handler,
        event: Event,
        api: API,
        ctx: Context,
    ) -> Result[typing.Any, BaseException]:
        try:
            return Ok(await run_handler(handler, event, ctx))
        except BaseException as exc:
            return await self.run(api, event, ctx, exc, handler.__name__)

    def match_exception(self, exception: BaseException) -> bool:
        for exc in self.exceptions:
            if isinstance(exc, type) and type(exception) is exc:
                return True
            if isinstance(exc, object) and type(exception) is type(exc):
                return True if not exc.args else exc.args == exception.args

        return False

    async def run(
        self,
        api: API,
        event: Event,
        ctx: Context,
        exception: BaseException,
        handler_name: str,
    ) -> Result[typing.Any, BaseException]:
        if self.match_exception(exception):
            logger.debug(
                "Error handler caught an exception {!r} in handler {!r}, running catcher {!r}...".format(
                    exception, handler_name, self.func.__name__
                )
            )
            return Ok(await run_handler(self.func, event, {"event": event, "api": api} | ctx))

        logger.debug("Failed to match exception {!r}.", exception.__class__.__name__)
        return Error(exception)


class ErrorHandler(ABCErrorHandler[Event]):
    def __init__(self, catcher: Catcher[Event] | None = None, /) -> None:
        self.catcher = catcher

    def __repr__(self) -> str:
        return (
            "<{}: exceptions_handled=[{}], catcher={!r}>".format(
                self.__class__.__name__,
                ", ".join(e.__name__ if isinstance(e, type) else repr(e) for e in self.catcher.exceptions),
                self.catcher,
            )
            if self.catcher is not None
            else "<{}()>".format(self.__class__.__name__)
        )

    def __call__(
        self,
        *exceptions: type[BaseException] | BaseException,
        logging: bool = False,
        raise_exception: bool = False,
        ignore_errors: bool = False,
    ):
        """Register the catcher.

        :param logging: Logging the result of the catcher at the level `DEBUG`.
        :param raise_exception: Raise an exception if the catcher has not started.
        :param ignore_errors: Ignore errors that may occur.
        """

        def decorator(func: F) -> F:
            if not self.catcher:
                self.catcher = Catcher(
                    func,
                    exceptions=list(exceptions),
                    logging=logging,
                    raise_exception=raise_exception,
                    ignore_errors=ignore_errors,
                )
            return func

        return decorator

    def _process_catcher_error(self, error: CatcherError) -> Result[None, BaseException]:
        assert self.catcher is not None

        if self.catcher.raise_exception:
            raise error.exc from None
        if self.catcher.logging:
            logger.error(error.message)
        if not self.catcher.ignore_errors:
            return Error(error.exc)

        return Ok(None)

    async def process(
        self,
        handler: Handler,
        event: Event,
        api: API,
        ctx: Context,
    ) -> Result[typing.Any, BaseException]:
        assert self.catcher is not None
        logger.debug("Processing the error handler for handler {!r}...", handler.__name__)

        try:
            return await self.catcher(handler, event, api, ctx)
        except BaseException as exc:
            return Error(
                CatcherError(
                    exc,
                    "Exception {} was occurred during the running catcher {!r}.".format(
                        repr(exc), self.catcher.func.__name__
                    ),
                )
            )

    async def run(
        self,
        handler: Handler,
        event: Event,
        api: API,
        ctx: Context,
    ) -> Result[typing.Any, BaseException]:
        if not self.catcher:
            return Ok(await run_handler(handler, event, ctx))

        match await self.process(handler, event, api, ctx):
            case Ok(value) as ok:
                if self.catcher.logging:
                    logger.debug(
                        "Catcher {!r} returned: {!r}",
                        self.catcher.func.__name__,
                        value,
                    )
                return ok
            case Error(exc) as err:
                if isinstance(exc, CatcherError):
                    return self._process_catcher_error(exc)
                if self.catcher.ignore_errors:
                    return Ok(None)
                return err


__all__ = ("Catcher", "ErrorHandler")



import dataclasses
import typing

import msgspec

from telegrinder.msgspec_utils import DataclassInstance, encoder
from telegrinder.types.objects import (
    CallbackGame,
    KeyboardButtonPollType,
    KeyboardButtonRequestChat,
    KeyboardButtonRequestUsers,
    LoginUrl,
    SwitchInlineQueryChosenChat,
    WebAppInfo,
)

KeyboardButton = typing.TypeVar("KeyboardButton", bound="BaseButton")


@dataclasses.dataclass
class BaseButton:
    def get_data(self) -> dict[str, typing.Any]:
        return {
            k: v if k != "callback_data" or isinstance(v, str) else encoder.encode(v)
            for k, v in dataclasses.asdict(self).items()
            if v is not None
        }


class RowButtons(typing.Generic[KeyboardButton]):
    buttons: list[KeyboardButton]
    auto_row: bool

    def __init__(self, *buttons: KeyboardButton, auto_row: bool = True) -> None:
        self.buttons = list(buttons)
        self.auto_row = auto_row

    def get_data(self) -> list[dict[str, typing.Any]]:
        return [b.get_data() for b in self.buttons]


@dataclasses.dataclass(slots=True)
class Button(BaseButton):
    text: str
    request_contact: bool = dataclasses.field(default=False, kw_only=True)
    request_location: bool = dataclasses.field(default=False, kw_only=True)
    request_chat: dict[str, typing.Any] | KeyboardButtonRequestChat | None = dataclasses.field(
        default=None, kw_only=True
    )
    request_user: dict[str, typing.Any] | KeyboardButtonRequestUsers | None = dataclasses.field(
        default=None, kw_only=True
    )
    request_poll: dict[str, typing.Any] | KeyboardButtonPollType | None = dataclasses.field(
        default=None, kw_only=True
    )
    web_app: dict[str, typing.Any] | WebAppInfo | None = dataclasses.field(default=None, kw_only=True)


@dataclasses.dataclass(slots=True)
class InlineButton(BaseButton):
    text: str
    url: str | None = dataclasses.field(default=None, kw_only=True)
    login_url: dict[str, typing.Any] | LoginUrl | None = dataclasses.field(default=None, kw_only=True)
    pay: bool | None = dataclasses.field(default=None, kw_only=True)
    callback_data: str | dict[str, typing.Any] | DataclassInstance | msgspec.Struct | None = (
        dataclasses.field(default=None, kw_only=True)
    )
    callback_game: dict[str, typing.Any] | CallbackGame | None = dataclasses.field(default=None, kw_only=True)
    switch_inline_query: str | None = dataclasses.field(default=None, kw_only=True)
    switch_inline_query_current_chat: str | None = dataclasses.field(default=None, kw_only=True)
    switch_inline_query_chosen_chat: dict[str, typing.Any] | SwitchInlineQueryChosenChat | None = (
        dataclasses.field(default=None, kw_only=True)
    )
    web_app: dict[str, typing.Any] | WebAppInfo | None = dataclasses.field(default=None, kw_only=True)


__all__ = (
    "BaseButton",
    "Button",
    "DataclassInstance",
    "InlineButton",
    "RowButtons",
)
